-- Initialization

GearEvaluator = LibStub("AceAddon-3.0"):NewAddon("GearEvaluator", "AceEvent-3.0", "AceConsole-3.0", "LibFuBarPlugin-3.0")

local slotnames = {"HeadSlot",
  "NeckSlot",
  "ShoulderSlot",
  "BackSlot",
  "ChestSlot",
  "ShirtSlot",
  "TabardSlot",
  "WristSlot",
  "HandsSlot",
  "WaistSlot",
  "LegsSlot",
  "FeetSlot",
  "Finger0Slot",
  "Finger1Slot",
  "Trinket0Slot",
  "Trinket1Slot",
  "MainHandSlot",
  "SecondaryHandSlot",
  "RangedSlot",
  "AmmoSlot",
  --"Bag0Slot",
  --"Bag1Slot",
  --"Bag2Slot",
  --"Bag3Slot"
}

-- Functions

function GearEvaluator:OnInitialize()
  if (not self.scores) then
    self.scores = {}
  end

  if (not self.errors) then
    self.errors = {}
  end

  if (not self.evaluators) then
    self.evaluators = {}
  end

  self.db = LibStub("AceDB-3.0"):New("GearEvaluatorDB")

  if (not self.db.profile.evaluators) then
    self.db.profile.evaluators = {}
  end

  self.evaluatorConfigTable = {
    name = "Evaluators",
	type = 'group',
	handler = self,
	childGroups = 'tab',
	args = {
	  add = {
        type = 'input',
        name = "Add evaluator",
		set = "AddEvaluatorFunction"
      }
    }
  }

  self.configTable = {
    name = "GearEvaluator",
    desc = "Evaluates your gear based on certain rules",
    type = 'group',
    args = {
	}
  }

  self:UpdateEvaluatorConfigTable()

  LibStub("AceConfig-3.0"):RegisterOptionsTable("GearEvaluator", self.configTable, { "gearevaluator" })
  LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("GearEvaluator Evaluators", self.evaluatorConfigTable)
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions("GearEvaluator")
  LibStub("AceConfigDialog-3.0"):AddToBlizOptions("GearEvaluator Evaluators", "Evaluators", "GearEvaluator")
end

function GearEvaluator:UpdateEvaluatorConfigTable()
  -- Delete non-existing entries
  for k, v in pairs(self.evaluatorConfigTable.args) do
    if (k ~= 'add' and not self.db.profile.evaluators[k]) then
      self.evaluatorConfigTable.args[k] = nil
    end
  end

  -- Add new entries
  for k, v in pairs(self.db.profile.evaluators) do
    self.evaluatorConfigTable.args[k] = {
      type = 'group',
	  name = k,
	  desc = "Options for "..k,
	  args = {
	    func = {
		  name = "Evaluator function",
		  type = 'input',
		  set = 'UpdateEvaluatorFunction',
		  get = 'GetEvaluatorFunction',
		  multiline = true,
		  width = 'full',
		  order = 0
		},
	    delete = {
	      type = 'execute',
	      name = "Delete",
	      desc = "Click to delete the "..k.." evaluator",
	      func = "DeleteEvaluatorFunction",
		  confirm = true,
	      confirmText = "You will not be able to undo this action!",
		  order = 1
	    }
	  }
	}
  end
end

function GearEvaluator:OnEnable()
  self:RegisterEvent("UNIT_INVENTORY_CHANGED")

  self:CompileAllEvaluators()
  self:UpdateScores()
end

function GearEvaluator:OnDisable()
  self:UnregisterEvent("UNIT_INVENTORY_CHANGED")
end

function GearEvaluator:AddEvaluatorFunction(info, name)
  if (not self.db.profile.evaluators[name]) then
    self:Print(name)
    self:Print("Adding evaluator "..name.."!")

    self.db.profile.evaluators[name] = "return 0"
    self.evaluators[name] = self:CompileEvaluator(self.db.profile.evaluators[name])
    self:UpdateScore(name)

    self:UpdateEvaluatorConfigTable()
  else
    self:Print("Evaluator "..name.." is already defined!")
  end
end

function GearEvaluator:DeleteEvaluatorFunction(info)
  local name = info[#info - 1]
  self:Print(name)

  self.db.profile.evaluators[name] = nil
  self.evaluators[name] = nil
  self.scores[name] = nil

  self:UpdateEvaluatorConfigTable()
end

function GearEvaluator:UpdateEvaluatorFunction(info, func)
  name = info[#info - 1]
  self.db.profile.evaluators[name] = func
  self.evaluators[name] = self:CompileEvaluator(func, name)
  self:UpdateScore(name)
end

function GearEvaluator:GetEvaluatorFunction(info)
  name = info[#info - 1]

  return self.db.profile.evaluators[name]
end

function GearEvaluator:CompileAllEvaluators()
  if (self.db.profile.evaluators) then
    for k, v in pairs(self.db.profile.evaluators) do
      self.evaluators[k] = self:CompileEvaluator(v, k)
    end
  end
end

function GearEvaluator:CompileEvaluator(text, name)
  local chunk, error = loadstring("return function(gear) "..text.." end", name)

  if (not chunk) then
    self:Print(error)
    self.errors[name] = error
  end

  return chunk()
end

function GearEvaluator:OnUpdateFuBarTooltip()
  GameTooltip:AddLine("Current Scores:")

  if (self.db.profile.evaluators) then
    for k in pairs(self.db.profile.evaluators) do
      GameTooltip:AddDoubleLine(k, self.scores[k] or self.errors[k])
    end
  end

  GameTooltip:AddLine("Left-click to modify evaluators")
  GameTooltip:AddLine("Right-click to edit preferences")
end

function GearEvaluator:GetScore(evalname, gear)
  if (not self.evaluators[evalname]) then
    return nil, "Undefined evaluator '"..evalname.."'"
  else
    local func = self.evaluators[evalname]
    return pcall(func, gear)
  end
end

function GearEvaluator:UpdateScore(name)
  local success, score = self:GetScore(name, self:GetCurrentEquipment())

  if (not success) then
    self:Print(score)
    self.scores[name] = nil
    self.errors[name] = score
  else
    self.scores[name] = score
  end
end

function GearEvaluator:UpdateScores()
  for k, v in pairs(self.evaluators) do
    self:UpdateScore(k)
  end
end

function GearEvaluator:GetCurrentEquipment()
  local gear = {}

  for k, s in pairs(slotnames) do
    gear[s] = GetInventoryItemLink("player", GetInventorySlotInfo(s))
  end

  return gear
end

-- Event handlers

function GearEvaluator:UNIT_INVENTORY_CHANGED(event, unit)
  if (unit == "player") then
    self:UpdateScores()
  end
end
